var OS = require('os')

var io = require('socket.io-client')
var SocketIoStream = require('socket.io-stream')

var Base = require('../base')
  , Crypto = require('./crypto')
  , HttpProxy = require('./http_proxy')
  , Logger = require('./logger')
  , PolyglotRelay = require('./polyglot_relay')
  , Publisher = require('./publisher')

var EVENTS = [ 'connect', 'connect_error', 'connect_timeout', 'reconnect', 'reconnect_error', 'reconnect_failed', 'error', 'disconnect', 'open', 'close', 'reconnect_attempt' ]

var debug = Base.logger('ix-connector')

var socket, proxyData
var isRelay = false
var isHttpProxy = false
var deviceData = {}

// Load .coffee
;(function(){
  var exp
  var mod = require('./connector-cs')
  for (exp in mod) { if (mod.hasOwnProperty(exp)) {
    exports[exp] = mod[exp]
  } }
})()

exports.connectToProxy = function(data){
  debug('connectToProxy', data)
  Logger.loadFilter(function(){
    Publisher.withDeviceData(function(err, devData){
      deviceData = devData
      proxyData = data
      if (Base.env.IX_PROXY_HOST_OVERRIDE) { proxyData.proxyUrl = "https://"+Base.env.IX_PROXY_HOST_OVERRIDE }
      console.log('Connecting to '+proxyData.proxyUrl+'...')
      socket = io(proxyData.proxyUrl, { reconnection: false, multiplex: false })
      setDebugHandlers(socket, EVENTS)
      setDebugHandlers(socket.io, EVENTS, 'manager:')
      socket.on('connect', onConnect)
      socket.on('connect_error', function(err){
        console.log('Connect error: '+err)
        exports.reconnect(isHttpProxy)
      })
      socket.on('disconnect', function(cause){
        console.log('Disconnected from '+proxyData.proxyUrl+': '+cause)
        exports.reconnect(isHttpProxy)
      })
      if (isHttpProxy) {
        var streamSocket = SocketIoStream(socket)
        setDebugHandlers(streamSocket, EVENTS, 'stream:')
        streamSocket.on('http-proxy', HttpProxy.onHttpProxy)
      }
      else {
        socket.on('ix-call', onIxCall)
        socket.on('ix-ping', onIxPing)
        socket.on('ix-x', onIxX)
      }
    })
  })
}

exports.connectRelayToProxy = function(data){
  debug('connectRelayToProxy', data)
  isRelay = true
  exports.connectToProxy(data)
}

exports.connectHttpToProxy = function(upstream, data){
  debug('connectHttpToProxy', upstream, data)
  isHttpProxy = true
  HttpProxy.setupHttpProxy(upstream)
  exports.connectToProxy(data)
}

var setDebugHandlers = function(socket, events, prefix){
  prefix = prefix || ''
  var i, event
  for (i = 0; i < events.length; i++) {
    event = events[i]
    socket.on(event, debugHandlerFor(prefix+event))
  }
}

var debugHandlerFor = function(type){
  return function(){
    var args = Array.prototype.slice.call(arguments, 0)
    args.unshift('event:'+type)
    debug.apply(this, args)
  }
}

var onConnect = function(){
  debug('connect')
  socket.emit('set-client', {
    apiConfig: Base.apiConfig(),
    proxyUrl: proxyData.proxyUrl,
    proxyExport: proxyData.proxyExport,
    device: deviceData,
    hostname: OS.hostname(),
    pid: process.pid,
    processKey: getProcessKey()
  }, function(err, data){
    debug('connect result', err, data)
    if (err) {
      console.log('Connect error: Unable to complete connection to '+proxyData.proxyUrl+': '+JSON.stringify(err))
      throw new Error(err)
    }
    else {
      console.log('Connected to '+proxyData.proxyUrl+'.')
    }
  })
}

var onIxCall = function(data, responder){
  debug('ix-call', arguments)
  try {
    Logger.logCallRequest(data)
    responder = Logger.wrapResponder(data, debug.wrap(responder))
  }
  catch(e){}
  try {
    if (isRelay) {
      Publisher.ixCall(data, responder, PolyglotRelay.invoker)
    }
    else {
      Publisher.ixCallNodejs(data, responder)
    }
  }
  catch(e) {
    responder(e)
  }
}

var onIxPing = function(replyData, responder){
  debug('ix-ping', replyData)
  responder(null, replyData)
}

var onIxX = function(cipherParams, responder){
  debug('ix-x', cipherParams)
  Crypto.decrypt(cipherParams, function(err, data){
    if (err) { responder.call(err); return }
    onIxCall(data, Crypto.encryptingResponder(responder))
  })
}

var PROCESS_KEY = null
var getProcessKey = function(){
  if (!PROCESS_KEY) {
    PROCESS_KEY = (new Date()).getTime()+'-'+Math.floor((Math.random()*1000000))
  }
  return PROCESS_KEY
}
